# Importation
import pygame as pg
from importation import importation_dossier

class Tuile(pg.sprite.Sprite):
    '''
    Classe de base pour la formation d'autre sprites héritant de cette classe
    '''
    def __init__(self,taille,pos):
        super().__init__() # Instruction nécessaire pour que la classe hérite des arguments de sprite.Sprite
        self.image = pg.Surface((taille,taille)) # Création d'une surface à partir de l'argument taille
        self.rect = self.image.get_rect(topleft = pos) # Méthode de la classe Surface qui permet de créer un rectangle associé

class TuileImmobile(Tuile):
    '''
    Classe pour les tuiles non animées : terre, ronces, racines, herbe, lianes...
    On rajoute un seul argument qui permet de charger des images avec la fonction importation_tuiles_coupees().
    Avec l'héritage, on évite de copier l'entièreté de la classe dont TuileImmobile hérite.
    '''
    
    def __init__(self,taille,pos,surface):
        super().__init__(taille,pos) 
        self.image = surface

class Champis(TuileImmobile):
    '''
    Classe spécifique aux champignons. Comme ce ne sont pas des tuiles à couper, on utilise la méthode load() pour créer une Surface à partir d'une image.
    Les champis sont trop hauts car chargés dans le coin supérieur gauche et pas toujours suffisamment grands pour atteindre le sol.
    '''

    def __init__(self,taille,pos,surface,decalage):
        super().__init__(taille,pos,surface) 
        self.image = pg.image.load(surface)
        y_decale = pos[1] - decalage
        self.rect = self.image.get_rect(topleft = (pos[0],y_decale))

    

class TuileAnimée(Tuile):
    '''
    Classe qui va charger une liste entière de surfaces grâce à la fonction importation_dossier() (voir importation) puis parcourir la liste grâce à un index.

    '''
    def __init__(self,taille,pos,chemin):
        super().__init__(taille,pos)
        self.images = importation_dossier(chemin) # Liste des surfaces
        self.index_images = 0
        self.image = self.images[self.index_images] 
        self.rect = self.image.get_rect(topleft = pos) # Le rectangle sera toujours aux dimensions déterminées par "taille" (ici 64 × 64) si je ne reprécise pas

    def animation(self):
        '''
        L'index est incrémenté à chaque frame et permet de changer d'image lorsque sa valeur entière change (on prend la prochaine image dans self.images)
        '''
        
        self.index_images += 0.14
        if self.index_images >= len(self.images): self.index_images = 0 # L'index est réinitialisé lorsqu'il atteint la longueur de la liste (dernière image affichée)
        self.image = self.images[int(self.index_images)]

    def update(self):
        '''
        Par défaut, cette méthode est vide. On peut y rajouter des méthodes que l'on veut effectuer à chaque frame. Ici, animation().
        La méthode update() est ensuite appelée dans la méthode disposition_niveau() de Niveau par l'intermédiaire de la méthode update() des groupes.
        '''
        self.animation()

class Arbre(TuileAnimée):
    def __init__(self,taille,pos,chemin,decalage):
        super().__init__(taille,pos,chemin)
        # L'arbre est trop bas car son point de placement est en haut à gauche de l'image dans un rectangle de 64 × 64. Il faut alors le remonter
        y_decale = pos[1] - decalage
        self.rect.topleft = (pos[0],y_decale)

    def animation(self): # Je modifie ici la méthode animation pour rendre l'animation plus lente et naturelle
        self.index_images += 0.09
        if self.index_images >= len(self.images): self.index_images = 0 
        self.image = self.images[int(self.index_images)]

class Lettre(TuileAnimée):
    def __init__(self,taille,pos,chemin):
        super().__init__(taille,pos,chemin)
        self.rect.center = (pos[0] + 32,pos[1] + 32) # Permet de confondre le centre du rectangle du sprite et le centre de la case où il se trouve. 32 est la moitié du côté d'une case.
        self.deroulement_texte = False

    def lecture_texte(self,joueur):
        '''
        Active la méthode evt_lettre() de Niveau lorsque le joueur rencontre pour la première fois la lettre.
        Autrement, la méthode evt_lettre() n'est pas appelée.
        '''

        if self.rect.colliderect(joueur.rect_collisions) and not self.deroulement_texte:
            # colliderect() vérifie si les deux rectangles se touchent et renvoie un booléen.
            # Cette instruction ne peut s'effectuer qu'une seule fois grâce au booléen.
            self.deroulement_texte = True
            return True
        else: return False

class Potion(TuileAnimée):
    def __init__(self,taille,pos,chemin):
        super().__init__(taille,pos,chemin)
        self.rect.center = (pos[0] + 32,pos[1] + 32)

    def animation(self): 
        self.index_images += 0.07
        if self.index_images >= len(self.images): self.index_images = 0 
        self.image = self.images[int(self.index_images)]
    
    def obtention(self,joueur):
        '''
        Active l'écran de fin si le joueur a une collision avec la potion.
        '''
        if self.rect.colliderect(joueur.rect_collisions):
            
            return True
        else: 
            return False


class Arrivee(TuileAnimée):
    '''
    Classe parente qui permet de définir l'indice en cours de choix de niveau. 
    On augmente l'indice en cours de 1 lorsqu'il y a une collision avec le joueur.
    Le rectangle de collision est un rectangle plus petit que l'image qui permet de gérer les collisions avec le joueur.
    '''
    def __init__(self,taille,pos,chemin,niveau_max):
        super().__init__(taille,pos,chemin)
        self.rect_collisions = self.rect
        self.arrive = False
        self.niveau_max = niveau_max
        self.niveau_en_cours = self.niveau_max

    def arrivee(self,joueur):
        '''
        Tant qu'il n'y a pas eu de collision, l'indice en cours est retourné. 
        Autrement, on prend un indice incrémenté de un pour prendre le prochain niveau et on retourne True pour activer le fondu.
        '''
        if self.rect_collisions.colliderect(joueur.rect_collisions) and not self.arrive:
            self.niveau_max += 1
            self.arrive = True

            return self.niveau_max,self.arrive

        else: return self.niveau_en_cours,False

# Les deux classes suivantes héritent de Arrivee. On les crée pour effectuer des décalages afin de placer l'image au bon endroit sur l'écran.
class Portail(Arrivee):
    def __init__(self,taille,pos,chemin,niveau_max):
        super().__init__(taille,pos,chemin,niveau_max)
        self.rect.y -= 16
        self.rect_collisions = pg.rect.Rect(pos[0] + 20,pos[1] + 60,24,20) # Deux premiers arguments : abscisse et ordonnée, deux derniers : largeur et hauteur.

class EntreeArbre(Arrivee):
    def __init__(self,taille,pos,chemin,niveau_max):
        super().__init__(taille,pos,chemin,niveau_max)
        self.rect = self.image.get_rect(bottomleft = (pos[0],pos[1] + 38))
        self.rect_collisions = pg.rect.Rect(pos[0] - 3,pos[1] - 128,self.rect.width,128)


class Bouton(TuileImmobile):
    '''
    Classe affichant des boutons au centre de l'écran à une certaine ordonnée.
    L'interaction avec ces boutons grâce au clic gauche de la souris permet d'effectuer diverses actions, précisées dans Niveau.
    '''
    def __init__(self,taille,pos,surface,surface_affichage):
        super().__init__(taille,pos,surface)
        self.image = pg.image.load(surface)
        self.rect = self.image.get_rect(center = pos)
        self.surface_affichage = surface_affichage

    def clic(self):
        '''
        Fonction qui gère le clic de la souris. Renvoie True si un clic a lieu.
        '''
        action = False
        pos_souris = pg.mouse.get_pos()
        clic = False

        if self.rect.collidepoint(pos_souris):
            # Si le rectangle du bouton a une collision par rapport à la position de la souris et si l'on clique une fois le clic gauche,
            # action est renvoyé True et permet d'effectuer justement des actions dans la classe Niveau

            if pg.mouse.get_pressed()[0] and not clic:
                action = True
                clic = True
        
        else: clic = False

        return action

    def affichage(self):
        self.surface_affichage.blit(self.image,self.rect) # Le deuxième argument permet de placer l'image selon le rect que l'on a défini.


class TuileAnimeePlus(pg.sprite.Sprite):
    '''
    Tuile animée avec plusieurs sets d'animations selon le statut du sprite.
    On vient prendre une liste dans un dictionnaire selon le string que porte self.statut (défini dans les classes filles), qui est changé sous certaines conditions.
    '''
    def __init__(self,pos):
        '''Je ne peux pas déjà appeler importation_statut() ici car les valeurs des variables définies dans __init__() changent à chaque classe'''
        super().__init__()
        self.pos = pos

        self.chemin_fichier = ""
        self.animations = {} 


    def importation_statut(self):
        '''
        Permet d'importer la liste des images correspondant à chacune des clés du dictionnaire.
        On ajoute au chemin initial le nom du dossier relatif à chaque set d'animations.
        '''
        
        for animation in self.animations.keys():
            chemin_complet = self.chemin_fichier + animation
            self.animations[animation] = importation_dossier(chemin_complet)
    
    def animation(self): pass # Méthode différente dans chaque classe fille. On veut juste la créer pour ensuite l'appeler dans update().

    def update(self): self.animation()

class Checkpoint(TuileAnimeePlus):
    '''
    Création d'objets qui permettent de redéfinir le point de réapparition du personnage lorsqu'il passe devant. 
    Lorsqu'il est mis KO, le joueur retourne non pas au début du niveau mais à ce point de réapparition.

    '''
    
    def __init__(self,pos):
        super().__init__(pos)

        # Animations et rectangle
        self.chemin_fichier =  "../Visuels/terrain/checkpoint/"
        self.animations = {"désactivé" : [],"activation" : [],"activé" : []} 
        self.importation_statut()

        self.index_images = 0
        self.image = self.animations["désactivé"][self.index_images]
        self.rect = self.image.get_rect(topleft = pos)
        self.rect.y -= 16

        # Statut
        self.statut = "désactivé"
        self.actif = False
        
    
    def animation(self):
        animation = self.animations[self.statut]

        self.index_images += 0.14
        if self.index_images >= len(animation): 
            if self.statut == "activation": self.statut = "activé" # Si le checkpoint atteint la fin de son animation d'activation, il passe au statut suivant.
            self.index_images = 0

        animation = self.animations[self.statut]
        
        self.image = animation[int(self.index_images)]



    def point_de_passage(self,pos_dep,joueur):
        '''
        Permet de redéfinir le point d'apparition du joueur lorsqu'il passe devant un checkpoint. La variable pos_dep est présente dans la classe Niveau.
        Lorsque le joueur passe devant le checkpoint, si ce dernier est le plus loin parmi ceux que le joueur a activés, la nouvelle position de départ est enregistrée.
        On effectue cette action qu'une seule fois en utilisant un booléen.
        '''

        if self.rect.colliderect(joueur.rect_collisions):
            if self.pos[0] > pos_dep[0] and not self.actif: 
                self.statut = "activation"
                self.index_images = 0
                pos_dep = self.pos
                self.actif = True

        return pos_dep


class ChampiRebond(TuileAnimeePlus):
    '''
    Classe qui définit un sprite permettant au joueur de rebondir s'il tombe dessus.
    L'action de rebond est codée dans la partie collisions verticales du joueur.
    '''
    def __init__(self,pos):
        super().__init__(pos)

        # Animations et rectangle
        self.chemin_fichier =  "../Visuels/terrain/champi_rbd/"
        self.animations = {"normal" : [],"utilisé" : []} 
        self.importation_statut()

        self.index_images = 0
        self.image = self.animations["normal"][self.index_images]
        self.rect = self.image.get_rect(topleft = pos)
        self.rect.y += 25

        # Statut
        self.statut = "normal"
    
    def animation(self):
        animation = self.animations[self.statut]

        self.index_images += 0.14
        if self.index_images >= len(animation): 
            self.index_images = 0
            if self.statut == "utilisé": self.statut = "normal"
        self.image = animation[int(self.index_images)]

        # Selon son état, le sprite a une taille d'image différente. Cette instruction permet de le replacer constamment au sol peu importe la taille de l'image.
        self.rect = self.image.get_rect(midbottom = self.rect.midbottom) 


class PlateformeCassante(TuileAnimeePlus):
    '''
    Plateforme particulière qui se brise en morceaux si le joueur marche dessus.
    Elle se reforme au bout d'un certain temps d'attente.
    L'interaction avec le joueur se fait aussi dans la classe Joueur.
    '''
    def __init__(self,pos):
        super().__init__(pos)

        # Animations et rectangle
        self.chemin_fichier =  "../Visuels/terrain/pltf_cass/"
        self.animations = {"normal" : [],"émiettement" : [],"reformation" : [],"cassée" : []} 
        self.importation_statut()

        self.index_images = 0
        self.image = self.animations["normal"][self.index_images]
        self.rect = self.image.get_rect(topleft = pos)

        # Statut
        self.statut = "normal"
        self.temps_attente = 10

    def animation(self):
        animation = self.animations[self.statut]

        self.index_images += 0.15
        if self.index_images >= len(animation):
            if self.statut == "reformation": self.statut = "normal"
            elif self.statut == "émiettement": self.statut = "cassée"
            self.index_images = 0

        animation = self.animations[self.statut] # Mise à jour de la liste et donc des animations (pas forcément nécessaire mais ici ça se voit)
        
        self.image = animation[int(self.index_images)]
        
    def reformation(self):
        '''
        Actions à effectuer selon le statut du sprite.
        '''

        # Si le sprite vient de se reformer mais que le temps n'a pas été réinitialisé, on le fait ici.
        # S'effectue uniquement lorsque le sprite se reforme après un KO (nécessaire si le point de réapparition est sur une de ces plateformes)
        if self.statut == "normal" and self.temps_attente <= 0: self.temps_attente = 10 
        elif self.statut == "cassée" and self.temps_attente > 0 or self.statut == "émiettement" and self.index_images >= 8:
            # Si la plateforme est entièrement ou suffisamment cassée, on supprime le rectangle de collision et on soustrait progressivement une valeur constante au temps d'attente.
            self.rect = pg.rect.Rect(self.pos[0],self.pos[1],0,0)
            self.temps_attente -= 0.05
        elif self.statut == "cassée" and self.temps_attente <= 0:
            # Si la plateforme est cassée et que le temps d'attente est fini, on réinitialise tous les paramètres du sprite.
            self.statut = "reformation"
            self.temps_attente = 10
            self.rect = self.image.get_rect(topleft = self.pos)


    def update(self):
        self.reformation()
        self.animation()


class FeuilleSaut(TuileAnimeePlus):
    '''
    Sprite sans collisions qui est affichée seulement lorsque le joueur double saute.
    Le sprite suit le rectangle du joueur, comme s'il était attaché à sa tête.
    '''
    def __init__(self,pos,groupe,joueur):
        super().__init__(pos)
        
        # Statut + ajout dans un groupe et sprite du joueur pour les interactions
        self.statut = "en_saut"

        self.joueur = joueur
        groupe.add(self)

        # Animations et rectangle
        self.chemin_fichier =  "../Visuels/feuille_saut/"
        self.animations = {"en_saut" : [],"chute" : [],"disparition" : []} 
        self.importation_statut()

        
        self.index_images = 0
        self.image = self.animations["en_saut"][self.index_images]
        self.rect = self.image.get_rect(bottomright = (self.joueur.rect.left + 18,self.joueur.rect.top + 24))



    def animation(self):
        animation = self.animations[self.statut]

        self.index_images += 0.15
        if self.index_images >= len(animation):
            self.index_images = 0
            if self.statut == "disparition": self.kill()

        animation = self.animations[self.statut] # Mise à jour de la liste
        
        self.image = animation[int(self.index_images)]

        # Le sprite suit l'orientation du joueur et le rectangle du joueur lui-même 
        if not self.joueur.orientation_droite: 
            self.image = pg.transform.flip(self.image,True,False) # Inverse l'image. Premier booléen : selon l'axe vertical, deuxième : selon l'axe horizontal.
            self.rect = self.image.get_rect(bottomleft = (self.joueur.rect.left + 16,self.joueur.rect.top + 24))
        else: self.rect = self.image.get_rect(bottomright = (self.joueur.rect.left + 18,self.joueur.rect.top + 24))

    def saut(self):
        '''
        Changement du statut du sprite sous certaines conditions.
        '''
        if self.joueur.direction.y < 0 and self.statut != "disparition": self.statut = "en_saut" # Si le joueur saute
        if self.joueur.direction.y > 1 and self.statut != "disparition": self.statut = "chute" # Si le joueur chute
        if self.statut != "disparition" and (self.joueur.sur_sol or self.joueur.KO or self.joueur.en_rebond): 
            # Si le joueur est sur le sol, est mis KO ou rebondit sur un champignon, la feuille disparaît.
            self.index_images = 0
            self.statut = "disparition"
            


    def update(self):
        self.animation()
        self.saut()

class Collectable(TuileAnimeePlus):
    '''
    Classe qui définit les sprites que le joueur peut en quelque sorte "collecter".
    En réalité, il disparaissent à son contact et permettent d'incrémenter une quelconque valeur ou d'effectuer une action.
    '''
    def __init__(self,pos,groupe):
        super().__init__(pos)

        # Animations et rectangle
        self.chemin_fichier = "../Visuels/collectable/"
        self.animations = {"sur_place" : [],"disparition" : []} 
        self.importation_statut()
        
        self.index_images = 0
        self.image = self.animations["sur_place"][self.index_images]
        self.rect = self.image.get_rect(center = (pos[0] + 32,pos[1] + 32)) # Centre le sprite

        # Statut et groupe
        self.statut = "sur_place"
        groupe.add(self)


    def animation(self):
        animation = self.animations[self.statut]

        # L'incrémentation de l'index est différente selon le set d'animations utilisé.
        if self.statut == "disparition": self.index_images += 0.1
        else: self.index_images += 0.09

        if self.index_images >= len(animation):
            self.index_images = 0
            if self.statut == "disparition": self.kill()

        animation = self.animations[self.statut] # Mise à jour de la liste
        
        self.image = animation[int(self.index_images)]

    def obtention(self,joueur):
        '''
        Permet d'effectuer une action avec le booléen lorsqu'il y a une collision avec le joueur et de faire disparaître le sprite.
        '''
        if self.rect.colliderect(joueur.rect_collisions):
            self.index_images = 0
            self.statut = "disparition"
            
            return True
        else: 
            return False

class FeuilleSautSurplace(Collectable):
    '''
    Collectable particulier. On ne change que le chemin pour les fichiers.
    '''
    def __init__(self,pos,groupe):
        super().__init__(pos,groupe)
        self.chemin_fichier = "../Visuels/feuille_saut/"
        self.importation_statut()
        self.rect = self.image.get_rect(topleft = pos)









